import arlunio as ar
import numpy as np

from .parameters import X, Y


@ar.definition
def Circle(x: X, y: Y, *, xc=0, yc=0, r=0.8, pt=None) -> ar.Mask:
    """
    .. arlunio-image::

        import arlunio as ar
        from arlunio.lib import Circle

        circle = Circle()
        image = ar.fill(circle(width=1920, height=1080))

    We define a circle using the following equality.

    .. math::

       (x - x_c)^2 + (y - y_c)^2 = r^2

    Attributes
    ----------
    xc:
        Corresponds with the :math:`x_c` variable in the equation above and defines the
        :math:`x`-coordinate of the circle's center.
    yc:
        Corresponds with the :math:`y_c` variable in the equation above and defines the
        :math:`y`-coordinate of the circle's center.
    r:
        Corresponds with the :math:`r` variable in the equation above and defines the
        radius of the circle.
    pt:
        If :code:`None`, then all points within the radius of the circle will be
        considered to be part of it. If this is set to some positive number then all
        points between radii :code:`(1 - pt) * r` and :code:`(1 + pt) * r` will be
        considered part of the circle.

    Examples
    --------
    Combining a few circles it's easy enough to draw a target

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle

       @ar.definition
       def Target(width: int, height: int):
           image = None
           parts = [
               (Circle(pt=0.02), "#000"),
               (Circle(r=0.75, pt=0.12), "#f00"),
               (Circle(r=0.6, pt=0.05), "#f00"),
               (Circle(r=0.4), "#f00"),
           ]

           for part, color in parts:
               image = ar.fill(
                   part(width=width, height=height), color=color, image=image
               )

           return image

       target = Target()
       image = target(width=1920, height=1080)

    Making use of the :code:`xc` and :code:`yc` attributes we can produce an
    approximation of the olympics logo

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle

       @ar.definition
       def OlympicRings(width: int, height: int, *, spacing=0.5, pt=0.025):

           dy = spacing / 4
           dx = spacing / 2
           args = {"scale": 0.5, "r": spacing, "pt": pt}

           image = None
           rings = [
               (Circle(yc=dy, xc=-(2.2 * dx), **args), "#0ff"),
               (Circle(yc=dy, **args), "#000"),
               (Circle(yc=dy, xc=(2.2 * dx), **args), "#f00"),
               (Circle(yc=-dy, xc=-(1.1 * dx), **args), "#ff0"),
               (Circle(yc=-dy, xc=(1.1 * dx), **args), "#0f0")
           ]

           for ring, color in rings:
               image = ar.fill(
                   ring(width=width, height=height), color=color, image=image
               )

           return image

       rings = OlympicRings()
       image = rings(width=1920, height=1080)
    """
    x = (x - xc) ** 2
    y = (y - yc) ** 2
    circle = np.sqrt(x + y)

    if pt is None:
        return circle < r ** 2

    inner = (1 - pt) * r ** 2
    outer = (1 + pt) * r ** 2

    return ar.all(inner < circle, circle < outer)


@ar.definition
def Ellipse(x: X, y: Y, *, xc=0, yc=0, a=2, b=1, r=0.8, pt=None) -> ar.Mask:
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import Ellipse

       ellipse = Ellipse()
       image = ar.fill(ellipse(width=1920, height=1080))

    An ellipse can be defined using the following equality.

    .. math::

       \\left(\\frac{x - x_c}{a}\\right)^2 +
       \\left(\\frac{y - y_c}{b}\\right)^2 = r^2

    Attributes
    ----------
    xc:
        Corresponds with the :math:`x_c` variable in the equation above and defines the
        :math:`x`-coordinate of the ellipse's center.
    yc:
        Corresponds with the :math:`y_c` variable in the equation above and defines the
        :math:`y`-coordinate of the ellipse's center.
    r:
        Corresponds with the :math:`r` variable in the equation above and controls the
        overall size of the ellipse.
    a:
        Corresponds with the :math:`a` variable in the equation above and controls the
        width of the ellipse.
    b:
        Corresponds with the :math:`b` variable in the equation above and controls the
        height of the ellipse.
    pt:
        If :code:`None` then all points within the radius of the ellipse will be
        considered to be part of it. If this is set to some positive number then all
        points between radii :code:`(1 - pt) * r` and :code:`(1 + pt) * r` will be
        considered part of the ellipse.

    Examples
    --------

    :code:`a` and :code:`b` together determine the overall shape of the ellipse.
    Increasing the value of :code:`a` will stretch the ellipse width wise, increasing
    :code:`b` has a similar effect for the height. It's worth noting that it's the
    ratio of these 2 values rather than their absolute values that has a greater effect
    on the shape of the ellipse. If :code:`a = b` then the equation simplifies to that
    of a circle

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Ellipse

       @ar.definition
       def EllipseDemo(width: int, height: int):
           image = None
           ellipses = [
               Ellipse(xc=-0.5, yc=-0.5, a=0.5, b=0.5, r=0.4),
               Ellipse(yc=-0.5, a=1, b=0.5, r=0.4),
               Ellipse(xc=0.5, yc=-0.5, a=2, b=0.5, r=0.4),
               Ellipse(a=1, b=1, r=0.4),
               Ellipse(xc=0.5, yc=0.5, a=2, b=2, r=0.4),
               Ellipse(xc=-0.5, a=0.5, b=1, r=0.4),
               Ellipse(xc=-0.5, yc=0.5, a=0.5, b=2, r=0.4)
           ]

           for ellipse in ellipses:
               image = ar.fill(ellipse(width=1920, height=1080), image=image)
           return image

       demo = EllipseDemo()
       image = demo(width=1920, height=1080)

    Playing around with the values and the coordinate inputs it's possible to draw
    something that looks like a diagram of an atom

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Ellipse, X, Y

       @ar.definition
       def Atom(x: X, y: Y):
           image = None

           ellipses = [
               (Ellipse(a=1.5, b=0.5, pt=0.005), x, y),
               (Ellipse(a=1.5, b=0.5, r=1, pt=0.005), x + y, y - x),
               (Ellipse(a=0.5, b=1.5, pt=0.005), x, y),
               (Ellipse(a=1.5, b=0.5, r=1, pt=0.005), x - y, x + y),
               (Ellipse(a=1, b=1, r=0.15), x, y)
           ]

           for ellipse, ex, ey in ellipses:
               image = ar.fill(ellipse(x=ex, y=ey), image=image)

           return image

       atom = Atom()
       image = atom(width=1920, height=1080)
    """

    x = (x - xc) ** 2
    y = (y - yc) ** 2

    a = a ** 2
    b = b ** 2

    ellipse = np.sqrt(x / a + y / b)

    if pt is None:
        return ellipse < r * r

    inner = (1 - pt) * r ** 2
    outer = (1 + pt) * r ** 2

    return ar.all(inner < ellipse, ellipse < outer)


@ar.definition
def SuperEllipse(
    x: X, y: Y, *, xc=0, yc=0, a=1, b=1, n=3, r=0.8, m=None, pt=None
) -> ar.Mask:
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import SuperEllipse

       ellipse = SuperEllipse()
       image = ar.fill(ellipse(width=1920, height=1080))

    We define a `SuperEllipse`_ by the following equality.

    .. math::

       \\left|\\frac{(x - x_c)}{a}\\right|^n + \\left|\\frac{(y - y_c)}{b}\\right|^m = r

    Attributes
    ----------
    xc:
        Corresponds with the :math:`x_c` variable in the equation above and defines the
        :math:`x`-coordinate of the center of the super ellipse.
    yc:
        Corresponds with the :math:`y_c` variable in the equation above and defines the
        :math:`y` -coordinate of the center of the super ellipse.
    r:
        Corresponds with the :math:`r` variable in the equation above and controls the
        size of the super ellipse.
    a:
        Corresponds with the :math:`a` variable in the equation above and controls the
        width of the super ellipse.
    b:
        Corresponds with the :math:`b` variable in the equation above and controls the
        height of the super ellipse.
    n:
        Corresponds with the :math:`n` variable in the equation above and controls the
        profile of the curve far from :math:`x = 0`
    m:
        Corresponds with the :math:`m` variable in the equation above and controls the
        profile of the curve close to :math:`x = 0`. If :code:`m = None` (default) then
        it will be set to the value of :code:`n`.
    pt:
        If :code:`None` then all points within the radius of the super ellipse will be
        considered to be part of it. If this is set to some positive number then all
        points between radii :code:`(1 - pt) * r` and :code:`(1 + pt) * r` will be
        considered part of the super ellipse.

    Examples
    --------

    Being a generalisation of the regular |Ellipse| definition most of the attributes
    will have a similar effect on the outcome so be sure to check it out for additional
    examples. For the :code:`SuperEllipse` definition the most interesting attributes
    are :code:`n` and :code:`m` greatly affect the shape of the super ellipse.

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import SuperEllipse

       @ar.definition
       def SuperEllipseDemo(width: int, height: int):
           image = None
           ellipses = [
               (SuperEllipse(n=0.5, pt=0.01),'#f00'),
               (SuperEllipse(n=1, pt=0.01),'#0f0'),
               (SuperEllipse(n=1.5, pt=0.01), '#00f'),
               (SuperEllipse(n=2, pt=0.01), '#ff0'),
               (SuperEllipse(n=3, pt=0.01), '#0ff')
           ]

           for ellipse, color in ellipses:
               image = ar.fill(
                   ellipse(width=1920, height=1080), color=color, image=image
               )

           return image

       demo = SuperEllipseDemo()
       image = demo(width=1920, height=1080)

    By default if you don't specify a value for :code:`m` it will inherit the value
    assigned to :code:`n`. However if you set :code:`m` to a different value then you
    can get even more interesting results!

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import SuperEllipse

       @ar.definition
       def Sauron(width: int, height: int):
           image = None
           ellipses = [
               (SuperEllipse(a=2, n=3, m=0.2, r=0.98),'#f00'),
               (SuperEllipse(n=2),'#f50'),
               (SuperEllipse(n=0.1, m=2), '#000'),
           ]

           for ellipse, color in ellipses:
               image = ar.fill(
                   ellipse(width=1920, height=1080), color=color, image=image
               )

           return image

       eye = Sauron()
       image = eye(width=1920, height=1080)

    .. _SuperEllipse: https://en.wikipedia.org/wiki/Superellipse

    """

    x = x - xc
    y = y - yc

    if m is None:
        m = n

    ellipse = np.abs(x / a) ** n + np.abs(y / b) ** m

    if pt is None:
        return ellipse < r

    inner = (1 - pt) * r
    outer = (1 + pt) * r

    return ar.all(inner < ellipse, ellipse < outer)


@ar.definition
def Empty(width: int, height: int) -> ar.Mask:
    """An empty shape definition.

    Example
    -------

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Empty

       e = Empty()
       image = ar.fill(e(width=1920, height=1080))
    """
    return np.full((height, width), False)


@ar.definition
def Full(width: int, height: int) -> ar.Mask:
    """An full shape definition.

    Example
    -------

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Full

       f = Full()
       image = ar.fill(f(width=1920, height=1080))
    """
    return np.full((height, width), True)


@ar.definition
def Square(x: X, y: Y, *, xc=0, yc=0, size=0.8, pt=None) -> ar.Mask:
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import Square

       square = Square()
       image = ar.fill(square(width=1920, height=1080))

    A square.

    Attributes
    ----------
    xc:
        Defines the :math:`x`-coordinate of the square's center
    yc:
        Defines the :math:`y`-coordinate of the square's center
    size:
        Defines the size of the square, sides will have a length of :code:`2 * size`
    pt:
        If :code:`None` then all points within the square's border will be considered to
        be part of it. If set to some positive number then all points within
        :code:`(1 - pt) * size` to :code:`(1 + pt) * size` of the border will be
        considered to be part of the square.

    Example
    -------

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Square, X, Y

       @ar.definition
       def SquareDemo(x: X, y: Y):
           image = None
           squares = [
               (Square(pt=0.01), x, y),
               (Square(pt=0.01), x + y, x - y),
               (Square(size=0.39, pt=0.01), x, y),
               (Square(size=0.39, pt=0.01), x + y, x - y),
               (Square(size=0.2), x, y),
           ]

           for square, sx, sy in squares:
               image = ar.fill(square(x=sx, y=sy), image=image)

           return image

       square = SquareDemo()
       image = square(width=1920, height=1080)
    """

    xs = np.abs(x - xc)
    ys = np.abs(y - yc)

    if pt is None:
        return ar.all(xs < size, ys < size)

    s = (1 - pt) * size
    S = (1 + pt) * size

    inner = ar.all(xs < s, ys < s)
    outer = ar.all(xs < S, ys < S)

    return ar.all(outer, ar.invert(inner))


@ar.definition
def Rectangle(x: X, y: Y, *, xc=0, yc=0, size=0.6, ratio=1.618, pt=None) -> ar.Mask:
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import Rectangle

       rectangle = Rectangle()
       image = ar.fill(rectangle(width=1920, height=1080))

    A Rectangle.

    Parameters
    ----------
    xc:
        Defines the :math:`x`-coordinate of the center of the rectangle
    yc:
        Defines the :math:`y`-coordinate of the center of the rectangle
    size:
        Defines the area of the rectangle
    ratio:
        Defines the ratio of the width to the height of the rectangle
    pt:
        If :code:`None` then all points within the rectangle's border will be considered
        to be part of it. If set to some positive number then all points within
        :code:`(1 - pt) * size` to :code:`(1 + pt) * size` of the border will be
        considered to be part of the rectangle.

    Examples
    --------

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Rectangle

       @ar.definition
       def RectangleDemo(width: int, height: int):
           image = None
           rects = [
               Rectangle(xc=-1, size=0.4, ratio=0.5),
               Rectangle(xc=0.25, yc=0.5, size=0.2, ratio=1),
               Rectangle(xc=0.5, yc=-0.5, size=0.4, ratio=2)
           ]

           for r in rects:
               image = ar.fill(r(width=width, height=height), image=image)

           return image

       demo = RectangleDemo()
       image = demo(width=1920, height=1080)
    """

    xs = np.abs(x - xc)
    ys = np.abs(y - yc)

    height = np.sqrt(size / ratio)
    width = height * ratio

    if pt is None:
        return ar.all(xs < width, ys < height)

    w, W = (1 - pt) * width, (1 + pt) * width
    h, H = (1 - pt) * height, (1 + pt) * height

    inner = ar.all(xs < w, ys < h)
    outer = ar.all(xs < W, ys < H)

    return ar.all(outer, ar.invert(inner))



import arlunio as ar

from arlunio import Defn, Mask


@ar.definition(operation=Defn.OP_ADD)
def MaskAdd(width: int, height: int, *, a: Defn[Mask] = None, b: Defn[Mask] = None):
    """Add any two mask producing definitions together.

    The resulting defintion will return a mask that is :code:`True` if a given point
    is :code:`True` in either :code:`a` or :code:`b`.

    Attributes
    ----------
    a:
        The first mask
    b:
        The second mask

    Examples
    --------
    This definition is used implicitly when adding two masks together

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle, Square

       c1 = Square(xc=-0.25, yc=-0.25, size=0.55)
       c2 = Circle(xc=0.25, yc=0.25, r=0.7)

       c = c1 + c2
       image = ar.fill(c(width=1920, height=1080))

    Or can be used directly

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle, MaskAdd, Square

       a = Square(xc=-0.25, yc=-0.25, size=0.55)
       b = Circle(xc=0.25, yc=0.25, r=0.7)

       c = MaskAdd(a=a, b=b)
       image = ar.fill(c(width=1920, height=1080))
    """

    return ar.any(a(width=width, height=height), b(width=width, height=height))


@ar.definition(operation=Defn.OP_SUB)
def MaskSub(width: int, height: int, *, a: Defn[Mask] = None, b: Defn[Mask] = None):
    """Subtract one mask away from another mask.

    The resulting definition will return a mask that is :code:`True` only if a given
    point is in :code:`a` **and not** in :code:`b`.

    .. note::

       You will get very different results depending on which way around you set
       :code:`a` and :code:`b`!

    Attributes
    ----------
    a:
        The first "base" mask
    b:
        The second mask that defines the region to remove from :code:`a`

    Examples
    --------
    This definition is used implicitly when subtracting one mask from another

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle, Square

       c1 = Square(xc=-0.25, yc=-0.25, size=0.55)
       c2 = Circle(xc=0.25, yc=0.25, r=0.7)

       c = c1 - c2
       image = ar.fill(c(width=1920, height=1080))

    Or can be used directly

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle, MaskSub, Square

       a = Square(xc=-0.25, yc=-0.25, size=0.55)
       b = Circle(xc=0.25, yc=0.25, r=0.7)

       c = MaskSub(a=b, b=a)
       image = ar.fill(c(width=1920, height=1080))
    """
    return ar.all(
        a(width=width, height=height), ar.invert(b(width=width, height=height))
    )


@ar.definition(operation=Defn.OP_MUL)
def MaskMul(width: int, height: int, *, a: Defn[Mask] = None, b: Defn[Mask] = None):
    """Muliply any two mask producing definitions together.

    The resulting definition will return a mask that is :code:`True` only if a given
    point is in both :code:`a` and :code:`b`.

    Attributes
    ----------
    a:
        The first mask
    b:
        The second mask

    Examples
    --------
    This definition is used implicitly when multiplying two masks together

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle, Square

       c1 = Square(xc=-0.25, yc=-0.25, size=0.55)
       c2 = Circle(xc=0.25, yc=0.25, r=0.7)

       c = c1 * c2
       image = ar.fill(c(width=1920, height=1080))

    Or can be used directly

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       from arlunio.lib import Circle, MaskMul, Square

       a = Square(xc=-0.25, yc=-0.25, size=0.55)
       b = Circle(xc=0.25, yc=0.25, r=0.7)

       c = MaskMul(a=a, b=b)
       image = ar.fill(c(width=1920, height=1080))
    """
    return ar.all(a(width=width, height=height), b(width=width, height=height))



import arlunio as ar
import numpy as np


@ar.definition
def X(width: int, height: int, *, x0=0, scale=1, stretch=False):
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import X

       x = X()
       image = ar.colorramp(x(width=1920, height=1080))

    Cartesian :math:`x` coordinates.

    Attributes
    ----------
    x0:
        Shift all the coordinate values by :code:`x0`
    scale:
        Controls the magnitude of the extreme values.
    stretch:
        If :code:`True` and the image is wider than it is tall then the grid will be
        stretched so that :code:`x = scale` falls on the image border. Otherwise the
        image's aspect ratio will be taken into account and :code:`x = scale` will fall
        somewhere within the boundaries of the image.

    Examples
    --------

    By default values will be generated between :math:`\\pm 1`::

       >>> from arlunio.lib import X
       >>> x = X()
       >>> x(width=4, height=4)
       array([[-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ]])

    If however the image is wider than it is tall this range will be extended so that
    the resulting image is not stretched::

       >>> x(width=4, height=2)
       array([[-2.        , -0.66666667,  0.66666667,  2.        ],
              [-2.        , -0.66666667,  0.66666667,  2.        ]])

    This behaviour can be disabled with the :code:`stretch` attribute::

       >>> x.stretch = True
       >>> x(width=4, height=2)
       array([[-1.        , -0.33333333,  0.33333333,  1.        ],
              [-1.        , -0.33333333,  0.33333333,  1.        ]])

    Additionally the :code:`scale` attribute can be used to adjust the magnitude of the
    extreme values generated, while the :code:`x0` attribute can be used to shift all
    the values by a given amount::

       >>> x = X(x0=-2, scale=2)
       >>> x(width=4, height=4)
       array([[0.        , 1.33333333, 2.66666667, 4.        ],
              [0.        , 1.33333333, 2.66666667, 4.        ],
              [0.        , 1.33333333, 2.66666667, 4.        ],
              [0.        , 1.33333333, 2.66666667, 4.        ]])

    """
    ratio = width / height

    if not stretch and ratio > 1:
        scale = scale * ratio

    x = np.linspace(-scale, scale, width)
    x = np.array([x for _ in range(height)])

    return x - x0


@ar.definition
def Y(width: int, height: int, *, y0=0, scale=1, stretch=False):
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import Y

       y = Y()
       image = ar.colorramp(y(width=1920, height=1080))

    Cartesian :math:`y` coordinates.

    Attributes
    ----------
    y0:
        Shift all the coordinate values by :code:`y0`
    scale:
        Controls the size of the extreme values
    stretch:
        If :code:`True` and the image is taller than it is wide then the grid will be
        stretched so that :code:`y = scale` falls on the border. Otherwise the image's
        aspect ratio will be taken into account and :code:`y = scale` will fall
        somewhere within the boundaries of the image.

    Examples
    --------

    By default values will be generated between :math:`\\pm 1`::

       >>> from arlunio.lib import Y
       >>> y = Y()
       >>> y(width=4, height=4)
       array([[ 1.        ,  1.        ,  1.        ,  1.        ],
              [ 0.33333333,  0.33333333,  0.33333333,  0.33333333],
              [-0.33333333, -0.33333333, -0.33333333, -0.33333333],
              [-1.        , -1.        , -1.        , -1.        ]])

    If however the image is taller than it is wide this range will be extended so that
    the resulting image is not stretched::

       >>> y(width=2, height=4)
       array([[ 2.        ,  2.        ],
              [ 0.66666667,  0.66666667],
              [-0.66666667, -0.66666667],
              [-2.        , -2.        ]])

    This behaviour can be disabled with the :code:`stretch` attribute::

       >>> y.stretch = True
       >>> y(width=2, height=4)
       array([[ 1.        ,  1.        ],
              [ 0.33333333,  0.33333333],
              [-0.33333333, -0.33333333],
              [-1.        , -1.        ]])

    Additionally the :code:`scale` attribute can be used to adjust the magnitude of the
    extreme values generated, while the :code:`y0` attribute can be used to shift all
    the values by a given amount::

       >>> y = Y(y0=-2, scale=2)
       >>> y(width=4, height=4)
       array([[4.        , 4.        , 4.        , 4.        ],
              [2.66666667, 2.66666667, 2.66666667, 2.66666667],
              [1.33333333, 1.33333333, 1.33333333, 1.33333333],
              [0.        , 0.        , 0.        , 0.        ]])

    """
    ratio = height / width

    if not stretch and ratio > 1:
        scale = scale * ratio

    y = np.linspace(scale, -scale, height)
    y = np.array([y for _ in range(width)]).transpose()

    return y - y0


@ar.definition
def R(x: X, y: Y):
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import R

       r = R()
       image = ar.colorramp(r(width=1920, height=1080))

    Polar :math:`r` coordinates.

    This definition corresponds with the distance a given point is from the origin and
    can be calculated from the point's equivalent Cartesian coordinate representation

    .. math::

       r = \\sqrt{x^2 + y^2}

    Examples
    --------
    ::

       >>> from arlunio.lib import R
       >>> r = R()
       >>> r(width=5, height=5)
       array([[1.41421356, 1.11803399, 1.        , 1.11803399, 1.41421356],
              [1.11803399, 0.70710678, 0.5       , 0.70710678, 1.11803399],
              [1.        , 0.5       , 0.        , 0.5       , 1.        ],
              [1.11803399, 0.70710678, 0.5       , 0.70710678, 1.11803399],
              [1.41421356, 1.11803399, 1.        , 1.11803399, 1.41421356]])

    While this definition does not currently have any attributes of its own, since it's
    derived from the |X| and |Y| definitions it automatically inherits the attributes
    from these base definitions::

       >>> r = R(x0=-2, y0=-2, scale=2)
       >>> r(width=5, height=5)
       array([[4.        , 4.12310563, 4.47213595, 5.        , 5.65685425],
              [3.        , 3.16227766, 3.60555128, 4.24264069, 5.        ],
              [2.        , 2.23606798, 2.82842712, 3.60555128, 4.47213595],
              [1.        , 1.41421356, 2.23606798, 3.16227766, 4.12310563],
              [0.        , 1.        , 2.        , 3.        , 4.        ]])

    Notice how in the case where the base definitions share a attribute (:code:`scale`
    in this case) they both share the value that is set when creating an instance of the
    :code:`R` definition.

    """
    return np.sqrt(x * x + y * y)


@ar.definition
def T(x: X, y: Y, *, t0=0):
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import T

       t = T()
       image = ar.colorramp(t(width=1920, height=1080))

    Polar, :math:`t` coordinates.

    This definition corresponds with the angle a given point is around from the positive
    :math:`x`-axis. This can be calculated from the point's equivalent Cartesian
    coordinate representation. All angles are given in :term:`radians`

    .. math::

       t = atan2\\left(\\frac{y}{x}\\right)

    Attributes
    ----------
    t0:
        Shift all the coordinate values by :code:`t0`

    Examples
    --------

    By default all point on the :math:`x`-axis will have a value of :code:`t = 0`::

       >>> from arlunio.lib import T
       >>> t = T()
       >>> t(width=5, height=5)
       array([[ 2.35619449,  2.03444394,  1.57079633,  1.10714872,  0.78539816],
              [ 2.67794504,  2.35619449,  1.57079633,  0.78539816,  0.46364761],
              [ 3.14159265,  3.14159265,  0.        ,  0.        ,  0.        ],
              [-2.67794504, -2.35619449, -1.57079633, -0.78539816, -0.46364761],
              [-2.35619449, -2.03444394, -1.57079633, -1.10714872, -0.78539816]])

    This can be changed however with the :code:`t0` attribute::

       >>> from math import pi
       >>> t.t0 = pi
       >>> t(width=5, height=5)
       array([[-0.78539816, -1.10714872, -1.57079633, -2.03444394, -2.35619449],
              [-0.46364761, -0.78539816, -1.57079633, -2.35619449, -2.67794504],
              [ 0.        ,  0.        , -3.14159265, -3.14159265, -3.14159265],
              [-5.8195377 , -5.49778714, -4.71238898, -3.92699082, -3.60524026],
              [-5.49778714, -5.17603659, -4.71238898, -4.24874137, -3.92699082]])

    Also, being a definition derived from |X| and |Y| the attributes for these
    definitions are also available to control the output::

       >>> t = T(x0=-2, y0=-2, scale=2)
       >>> t(width=5, height=5)
       array([[1.57079633, 1.32581766, 1.10714872, 0.92729522, 0.78539816],
              [1.57079633, 1.24904577, 0.98279372, 0.78539816, 0.64350111],
              [1.57079633, 1.10714872, 0.78539816, 0.5880026 , 0.46364761],
              [1.57079633, 0.78539816, 0.46364761, 0.32175055, 0.24497866],
              [0.        , 0.        , 0.        , 0.        , 0.        ]])

    """
    t = np.arctan2(y, x)
    return t - t0



import base64
import enum
import io
import logging
import pathlib

from typing import Optional

import numpy as np
import PIL.Image as PImage
import PIL.ImageColor as PColor

from ._expressions import lerp, normalise

logger = logging.getLogger(__name__)

# Create a type alias that we're free to change in the future
Image = PImage.Image


class Resolutions(enum.Enum):
    """Enum that defines some common image resolutions

    Members of this enum are tuples containing the width and height which can be
    accessed by name::

       >>> from arlunio import Resolutions as R

       >>> hd = R.HD
       >>> hd.width
       1280

       >>> hd.height
       720

    Resolutions can also unpacked::

       >>> width, height = hd
       >>> width
       1280

       >>> height
       720
    """

    HD = (1280, 720)
    """1280 x 720"""

    FHD = (1920, 1080)
    """1920 x 1080"""

    QHD = (2560, 1440)
    """2560 x 1440"""

    def __iter__(self):
        value = self.value
        return iter([value[0], value[1]])

    @property
    def width(self):
        return self.value[0]

    @property
    def height(self):
        return self.value[1]


def save(image, filename: str, mkdirs: bool = False) -> None:
    """Save an image in PNG format.

    :param filename: The filepath to save the image to.
    :param mkdirs: If true, make any parent directories
    """
    path = pathlib.Path(filename)

    if not path.parent.exists() and mkdirs:
        path.parent.mkdir(parents=True)

    with open(filename, "wb") as f:
        image.save(f)


def encode(image) -> bytes:
    """Return the image encoded as a base64 string."""

    with io.BytesIO() as byte_stream:
        image.save(byte_stream, "PNG")
        image_bytes = byte_stream.getvalue()

        return base64.b64encode(image_bytes)


def colorramp(values, start: Optional[str] = None, stop: Optional[str] = None) -> Image:
    """Given a range of values, produce an image mapping those values onto colors."""

    vs = normalise(values)
    (r, g, b) = PColor.getrgb("#000") if start is None else PColor.getrgb(start)
    (R, G, B) = PColor.getrgb("#fff") if stop is None else PColor.getrgb(stop)

    reds = np.floor(lerp(r, R)(vs))
    greens = np.floor(lerp(g, G)(vs))
    blues = np.floor(lerp(b, B)(vs))

    pixels = np.array(np.dstack([reds, greens, blues]), dtype=np.uint8)
    return PImage.fromarray(pixels)


def fill(
    mask,
    color: Optional[str] = None,
    background: Optional[str] = None,
    image: Optional[Image] = None,
) -> Image:
    """Given a mask, fill it in with a color.

    Parameters
    ----------
    mask:
        The mask used to select the pixels to fill in
    color:
        A string representation of the color to use, this can be in any format that is
        supported by Pillow's |PIL.ImageColor| module. If omitted this will default to
        black.
    background:
        In the case where an existing image is not provided this parameter can be used
        to set the background color of the generated image. This can be any string that
        is accepted by the |PIL.ImageColor| module. If omitted this will default to
        white.
    image:
        The image to color in, if omitted a new image will be generated.

    Returns
    -------
    Image
        An image with the region selected by the mask colored with the given color

    """

    color = "#000" if color is None else color
    fill_color = PColor.getrgb(color)

    mask_img = PImage.fromarray(mask)

    if image is None:
        background = "#fff" if background is None else background

        height, width = mask.shape
        image = PImage.new("RGB", (width, height), color=background)

    else:
        image = image.copy()

    image.paste(fill_color, mask=mask_img)

    return image



import inspect
import textwrap

from typing import Any, Dict, List, Optional

import arlunio

from sphinx.application import Sphinx

from .builder import NotebookTutorialBuilder
from .directives import (
    ArlunioImageDirective,
    NBTutorialDirective,
    depart_nbtutorial,
    nbtutorial,
    visit_nbtutorial,
)

# fmt: off
TEMPLATE = [
    ".. list-table::",
    "   :widths: 5 20",
    "   :header-rows: 0",
    "",
]
# fmt: on


def _document_inheritance(defn: arlunio.Defn) -> Optional[List[str]]:
    """Given a definition, link back to any definitions it derives from."""

    lines = []
    inputs = []
    defns = []

    for name, value in defn.inputs().items():
        dtype = value.dtype

        if hasattr(value.dtype, "__name__"):
            dtype = dtype.__name__

        inputs.append(f":code:`{name}: {dtype}`")

    if len(inputs) > 0:
        lines.append(textwrap.indent("* - **Inputs:**", " " * 3))
        lines.append(textwrap.indent(", ".join(inputs), " " * 5 + "- "))

    for name, value in defn.bases().items():

        name = value.__name__
        mod = value.__module__

        defns.append(f":class:`{name} <{mod}.{name}>`")

    if len(defns) > 0:
        lines.append(textwrap.indent("* - **Bases:**", " " * 3))
        lines.append(textwrap.indent(", ".join(defns), " " * 5 + "- "))

    if len(lines) == 0:
        return None

    return list(TEMPLATE) + lines + ["", ""]


def _process_docstring(
    app: Sphinx, what: str, name: str, obj: Any, options: Any, lines: List[str]
) -> None:

    if inspect.isclass(obj) and issubclass(obj, arlunio.Defn):
        inherits = _document_inheritance(obj)

        if inherits is not None:

            for l in reversed(inherits):
                lines.insert(0, l)


def setup(app: Sphinx) -> Dict[str, Any]:

    app.add_node(
        nbtutorial,
        html=(visit_nbtutorial, depart_nbtutorial),
        latex=(visit_nbtutorial, depart_nbtutorial),
        text=(visit_nbtutorial, depart_nbtutorial),
    )
    app.add_builder(NotebookTutorialBuilder)

    app.setup_extension("sphinx.ext.autodoc")
    app.connect("autodoc-process-docstring", _process_docstring)

    app.add_directive("nbtutorial", NBTutorialDirective)
    app.add_directive("arlunio-image", ArlunioImageDirective)

    return {"version": arlunio.__version__, "parallel_read_safe": True}



from .operators import MaskAdd, MaskMul, MaskSub  # noqa: F401
from .parameters import R, T, X, Y  # noqa: F401
from .pattern import Grid, Map, Pixelize  # noqa: F401
from .shapes import (  # noqa: F401
    Circle,
    Ellipse,
    Empty,
    Full,
    Rectangle,
    Square,
    SuperEllipse,
)



import logging

import arlunio as ar
import numpy as np

logger = logging.getLogger(__name__)


@ar.definition
def Grid(width: int, height: int, *, n=4, m=None, defn=None) -> ar.Mask:
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import Circle, Grid

       pattern = Grid(defn=Circle())
       image = ar.fill(pattern(width=1920, height=1080))

    Repeatedly draw the given defintition in a grid.

    This works by splitting the resolution of the final image into an :code:`n` by
    :code:`m` grid and rendering the given definition instance once and tiling the
    result across the remaining grid squares.

    .. note::

       Due to a limitation in the current implementation of this, you will get the best
       results if your :code:`n` by :code:`m` grid divides cleanly into the resolution
       of the final image.

    Attributes
    ----------
    n:
        The number of times to repeat the shape horizontally across the image
    m:
        The number of times to repeat the shape vertically across the image. If
        :code:`None` this defaults to the value of :code:`n`
    defn:
        The instance of the definition to draw.

    Examples
    --------
    Pattern generated from circles

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       import numpy as np

       from arlunio.lib import Circle, Grid, X, Y

       @ar.definition
       def Template(x:X, y: Y):
           c = Circle(xc=.4, yc=.4, pt=.02)
           return c(x=np.abs(x), y=np.abs(y))

       pattern = Grid(defn=Template(scale=1.))
       image = ar.fill(
           pattern(width=1080, height=1080), background="#000", color="#ff0"
       )

    A checkerboard like pattern

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       import numpy as np

       from arlunio.lib import Grid, X, Y

       @ar.definition
       def Template(x: X, y: Y):
           return np.abs(x) - np.abs(y) < 0

       grid = Grid(defn=Template(), n=22, m=13)
       image = ar.fill(grid(width=1920, height=1080))
    """
    if m is None:
        m = n

    bg = np.full((height, width), False)

    # Draw the shape at a size determined by the size of the grid
    s_height, s_width = height // m, width // n
    mask = defn(width=s_width, height=s_height)

    # Let numpy handle the repeating of the shape across the image.
    pattern = np.tile(mask, (m, n))

    # Apply the pattern to the background, depending on the grid size and
    # image dimensions align, the generated grid may not perfectly fill the
    # image.
    p_height, p_width = pattern.shape

    bg[:p_height, :p_width] = pattern

    return bg


@ar.definition
def Map(width: int, height: int, *, layout=None, legend=None) -> ar.Mask:
    """For more complex layouts.

    .. note::

       Due to a limitation in the current implementation, you will get best results
       if the dimenions of the :code:`layout` grid divide cleanly into the dimensions
       of the final image.

    Attributes
    ----------
    layout:
        A 2D numpy array of values detailing what to draw where.
    legened:
        A dictionary that maps values from the layout to definition instances to draw

    Example
    -------
    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       import numpy as np

       from arlunio.lib import Empty, Map, Rectangle

       @ar.definition
       def Wall(width: int, height: int, *, sides=None):
           r = 50
           d = 1
           walls = {
               'top': {"yc": d, "ratio": r},
               'bottom': {"yc": -d, "ratio": r},
               'left': {"xc": -d, "ratio": 1/r},
               "right": {"xc": d, "ratio": 1/r}
           }

           mask = False
           for side in sides.split('-'):
               wall = Rectangle(size=0.2, **walls[side])
               mask = ar.any(mask, wall(width=width, height=height))

           return mask

       legend = {
           "": Empty(),
           "tt": Wall(sides="top"),
           "bb": Wall(sides="bottom"),
           "ll": Wall(sides="left"),
           "rr": Wall(sides="right"),
           "tl": Wall(sides="top-left"),
           "tr": Wall(sides="top-right"),
           "bl": Wall(sides='bottom-left'),
           "br": Wall(sides="bottom-right")
       }

       layout = np.array([
           ["tt", "tt", "tt", "tt", "tr"],
           [  "", "tl", "tt", "tr", "rr"],
           [  "", "ll", "bl", "br", "rr"],
           [  "", "bl", "bb", "bb", "br"],
           [  "",   "",   "",   "",   ""]
       ])

       map_ = Map(legend=legend, layout=layout)
       image = ar.fill(map_(width=1080, height=1080), color="blue")
    """

    # TODO: Handle divisions with rounding errors
    nx, ny = len(layout), len(layout[0])
    size = {"height": height // ny, "width": width // nx}

    # Build a new dict with the values being the shapes drawn at the appropriate res
    # to ensure we only draw them once.
    items = {k: v(**size) for k, v in legend.items()}
    return np.block([[items[key] for key in row] for row in layout])


@ar.definition
def Pixelize(
    width: int, height: int, *, pixels=None, defn=None, n=None, m=None
) -> ar.Mask:
    """
    .. arlunio-image::

       import arlunio as ar
       from arlunio.lib import Circle, Pixelize

       pix = Pixelize(defn=Circle(), n=32, m=32)
       image = ar.fill(pix(width=1920, height=1080))

    Draw a pixelated version of a definition.

    .. note::

       Due to a limitation in the current implementation you will get the best results
       if your values for :code:`n` and :code:`m` cleanly divides into the resolution of
       the final image. This also applies to the dimensions of your array if you are
       providing the :code:`pixels` attribute directly.

    .. note::

       Due to how each enlarged pixel in the final image is mapped onto a rectangular
       region of real pixels, unless your :code:`n` by :code:`m` grid matches the aspect
       ratio of the final image you will find your original shape will be distorted as
       it is strectched to cover the final image.

    Attributes
    ----------
    pixels:
        A 2D boolean numpy array that specifies which pixels to color in.
    defn:
        An instance of a definition to use to generate the grid of pixels from.
    n:
        Required when specifying the :code:`defn` attribute, sets the number of pixels
        to generate horizontally
    m:
        Required when specifying the :code:`defn` attribute, set the number of pixels
        to generate vertically

    Examples
    --------
    This definition can be used to render a simple pixel pattern at a higher resolution

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       import numpy as np

       from arlunio.lib import Pixelize

       pixels = np.array([
           [False,  True,  True, False],
           [ True, False, False,  True],
           [ True, False, False,  True],
           [False,  True,  True, False]
       ])
       defn = Pixelize(pixels=pixels)
       image = ar.fill(defn(width=1080, height=1080))

    Alternatively we can generate the pixels from an instance of another definition

    .. arlunio-image::
       :include-code: before

       import arlunio as ar
       import numpy as np

       from arlunio.lib import Circle, X, Y, Pixelize

       @ar.definition
       def Ghost(x: X, y: Y):
           head = Circle(yc=0.5, r=0.7)
           eyes = Circle(xc=0.2, yc=0.6, r=0.3)

           body = ar.all(
               y < 0.5,
               np.abs(x) < 0.49,
               0.1 * np.cos(5 * np.pi * x) - 0.3 < y
           )

           return ar.any(
               ar.all(
                   head(x=x, y=y),
                   ar.invert(eyes(x=np.abs(x), y=y))
               ),
               body
           )

       ghost = Pixelize(defn=Ghost(y0=-0.3), n=32, m=32)
       image = ar.fill(ghost(width=1080, height=1080), color="#f00")
    """

    if defn is None and pixels is None:
        raise ValueError("You must either provide a shape or a pixel pattern.")

    if defn is not None:

        if n is None or m is None:
            raise ValueError("You must also provide the `n` and `m` attributes")

        pixels = defn(width=n, height=m)

    n, m = len(pixels), len(pixels[0])
    size = (height // m, width // n)  # TODO: Handle divisions with rounding errors

    fill = np.full(size, True)
    empty = np.full(size, False)

    return np.block([[fill if col else empty for col in row] for row in pixels])



"""Import utilities.

This module defines a number of classes that allow us to extend Python's import
machinery to support more than just regular Python files. Currently this includes
support for the following formats.

- .ipynb - Jupyter Notebooks

Note: The code that adds support for Jupyter Notebooks is heavily based on code
originally found in the documentation.

https://jupyter-notebook.readthedocs.io/en/stable/examples/Notebook/Importing%20Notebooks.html
"""
import importlib.util as imutil
import logging
import os
import pathlib

from importlib.abc import Loader, MetaPathFinder

import nbformat

from IPython.core.interactiveshell import InteractiveShell

logger = logging.getLogger(__name__)


def _find_notebook(fullname, path=None):
    """Converts a.b.c into a valid filepath and checks to see if it exists."""

    name = fullname.rsplit(".", 1)[-1]

    if path is None:
        path = [""]

    if isinstance(path, str):
        path = [path]

    logger.debug("path: %s", path)

    for dir in path:
        nb_path = os.path.join(dir, name + ".ipynb")
        logger.debug("Trying %s", nb_path)

        if os.path.isfile(nb_path):
            return nb_path

        nb_path = nb_path.replace("_", " ")
        logger.debug("Trying %s", nb_path)

        if os.path.isfile(nb_path):
            return nb_path


class NotebookLoader(Loader):
    """This does the legwork of importing a notebook.

    Note: It's not currently clear when all the various loaders should be used so we're
    sticking with a vanilla Loader for now.
    """

    def __init__(self, path=None):
        self.path = path
        self.shell = InteractiveShell.instance()
        self.shell.enable_gui = lambda x: False

    def exec_module(self, module):
        module.__file__ = _find_notebook(module.__name__, self.path)

        with open(module.__file__, "r", encoding="utf-8") as f:
            # Read notebook using v4 of the spec
            notebook = nbformat.read(f, 4)
            module.__notebook__ = notebook

        for cell in notebook.cells:
            if cell.cell_type == "code":

                # Transform any magics into python code.
                src = cell.source
                code = self.shell.input_transformer_manager.transform_cell(src)
                exec(code, module.__dict__)

    @classmethod
    def fromfile(cls, filepath: str):
        """Import a Jupyter Notebook directly from a filepath."""
        nbpath = pathlib.Path(filepath)

        if not nbpath.exists():
            raise FileNotFoundError(nbpath)

        nbname = nbpath.stem.replace(" ", "_")

        loader = cls(str(nbpath.parent))
        spec = imutil.spec_from_file_location(nbname, str(nbpath), loader=loader)

        module = imutil.module_from_spec(spec)
        spec.loader.exec_module(module)

        return module


class NotebookFinder(MetaPathFinder):
    """Responsible for checking a path to see if it "looks like a notebook".

    If we think we can handle a given path then we should return a loader capable of
    processing it.

    Note: It's not currently clear what the difference is between MetaPathFinder and
    PathEntryFinder, currently implementing the MetaPathFinder interface since it seems
    to work correctly.
    """

    def __init__(self):
        self.loaders = {}

    def find_spec(self, fullname, path, target=None):

        nb_path = _find_notebook(fullname, path)
        if not nb_path:
            return

        key = os.path.sep.join(path) if path else path
        logger.debug("key: %s", key)

        if key not in self.loaders:
            self.loaders[key] = NotebookLoader(path)

        loader = self.loaders[key]
        return imutil.spec_from_loader(fullname, loader)



import string
import textwrap
import traceback

from typing import List

import arlunio as ar

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst.directives.admonitions import BaseAdmonition
from docutils.statemachine import StringList
from PIL.Image import Image
from sphinx.util import logging, nested_parse_with_titles

logger = logging.getLogger(__name__)


ERROR_TEMPLATE = """\
.. error::

${message}

${traceback}
"""

PREVIEW_TEMPLATE = """\
.. raw:: html

   <figure>
     <img alt="${alt_tag}"
          src="data:image/png;base64,${img_data}"
          style="border: solid 1px #ddd"/>
     <figcaption style="margin: 10px;text-align:center">
       <small style="font-style:italic">${alt_tag}</small>
     </figcaption>
   </figure>

"""

IMAGE_TEMPLATE = string.Template(
    """\
<figure style="border: solid 1px #ddd;margin:auto">
  <img style="image-rendering:$rendering;width:100%"
       src="data:image/png;base64,$data"></img>
</figure>
"""
)


class nbtutorial(nodes.General, nodes.Element):
    pass


def visit_nbtutorial(self, node):
    pass


def depart_nbtutorial(self, node):
    pass


def format_error(message: str, err: str) -> StringList:
    """Given an error message format it as an error."""

    indent = " " * 3
    msg = textwrap.indent(message + "::", indent)
    err_msg = textwrap.indent(err, indent * 2)

    template = string.Template(ERROR_TEMPLATE)
    error = template.safe_substitute({"message": msg, "traceback": err_msg})

    return StringList(error.split("\n"), source="")


def parse_content(state, content: StringList) -> List[nodes.Node]:
    """Given a reStructured text representation of some content format parse it
    into a list of nodes."""

    section = nodes.section()
    section.document = state.document
    nested_parse_with_titles(state, content, section)

    return section.children


def render_image(src: str, smooth: bool = True) -> List[nodes.Node]:
    """Given the source code for an image return a doctree that when rendered by
    Sphinx will insert that image into a HTML page.

    Parameters
    ----------
    src:
        The source code that produces the image.
    smooth:
        If :code:`True` (default) allow the browser to scale the image using an
        algorithm that smooths out the edges of the image.
    """
    doctree = []

    try:
        code = compile(src, "<string>", "exec")
    except Exception:
        message = nodes.Text("Unable to render image: Invalid code")
        err = nodes.literal_block("", traceback.format_exc())
        doctree.append(nodes.error("", message, err))

        return doctree

    environment = {}

    try:
        exec(code, environment)
    except Exception:
        message = nodes.Text("Unable to render image: Error in code")
        err = nodes.literal_block("", traceback.format_exc())
        doctree.append(nodes.error("", message, err))

        return doctree

    image = None

    for obj in environment.values():
        if isinstance(obj, Image):
            image = obj

    if image is not None:
        context = {
            "data": ar.encode(image).decode("utf-8"),
            "rendering": "auto" if smooth else "crisp-edges",
        }
        html = IMAGE_TEMPLATE.safe_substitute(context)
        doctree.append(nodes.raw("", html, format="html"))

    return doctree


class NBTutorialDirective(rst.Directive):
    def run(self):
        return [nbtutorial("")]


class nbsolution(nodes.General, nodes.Element):
    pass


def visit_nbsolution(self, node):
    self.body.append('<details class="admonition note">\n')
    self.body.append(
        '<summary class="admonition-title">Solution (click to expand)</summary>\n'
    )


def depart_nbsolution(self, node):
    self.body.append("</details>\n")


class NBSolutionDirective(BaseAdmonition):

    has_content = True
    node_class = nbsolution

    def run(self):
        (soln,) = super().run()
        return [soln]


class ArlunioImageDirective(rst.Directive):
    """Given some code that produces an image, render it in the page."""

    has_content = True

    option_spec = {
        "include-code": rst.directives.unchanged,
        "disable-smoothing": rst.directives.flag,
    }

    def run(self):

        src = "\n".join(self.content)
        smooth = not ("disable-smoothing" in self.options.keys())
        nodelist = render_image(src, smooth)

        if "include-code" in self.options.keys():

            code_block = nodes.literal_block("", src)
            code_block["language"] = "python"

            if self.options["include-code"] == "before":
                nodelist.insert(0, code_block)
            else:
                nodelist.append(code_block)

        return nodelist



import inspect
import logging

from typing import Any, ClassVar, Dict, List, Optional

import attr

logger = logging.getLogger(__name__)


class Mask:
    """Currently just a type alias for boolean numpy arrays but gives us the flexibility
    to add smarts later."""


def _format_type(obj: Optional[Any] = None, type_: Optional[Any] = None) -> str:
    """Given an object, return an appropriate representation for its type."""

    if obj is not None and type_ is not None:
        raise ValueError("You can only specify 'obj' or 'type_'")

    if obj is not None and isinstance(obj, Defn):
        type_ = obj.produces()

        try:
            t = type_.__name__
        except AttributeError:
            t = str(type_).split(".")[-1]

        return f"{obj.__class__.__name__}[{t}]"

    if obj is not None:
        type_ = type(obj)

    try:
        return type_.__name__
    except AttributeError:
        return str(type_).split(".")[-1]


@attr.s(auto_attribs=True, repr=False, frozen=True)
class DefnSignature:
    """A class that represents the 'function signature' of a definition."""

    produces: Any
    """The type of object the definition produces."""

    def __repr__(self):
        t = _format_type(type_=self.produces)
        return f"Defn[{t}]"


class _BaseDefn(type):
    """A metaclass for prodviding a few goodies on the definition class itself."""

    def __getitem__(self, key):
        return DefnSignature(produces=key)


@attr.s(auto_attribs=True)
class Defn(metaclass=_BaseDefn):
    """Defn, short for 'Definition' is the object that powers the rest of
    :code:`arlunio`.
    """

    ATTR_ID: ClassVar[str] = "arlunio.attribute"

    OP_ADD: ClassVar[str] = "addition"
    OP_AND: ClassVar[str] = "and"
    OP_DIV: ClassVar[str] = "division"
    OP_FLOORDIV: ClassVar[str] = "floor_division"
    OP_LSHIFT: ClassVar[str] = "left_shift"
    OP_MATMUL: ClassVar[str] = "matrix_multiplication"
    OP_MOD: ClassVar[str] = "modulus"
    OP_MUL: ClassVar[str] = "multiplication"
    OP_OR: ClassVar[str] = "or"
    OP_POW: ClassVar[str] = "power"
    OP_RSHIFT: ClassVar[str] = "right_shift"
    OP_SUB: ClassVar[str] = "subtraction"
    OP_XOR: ClassVar[str] = "exclusive_or"

    def __call__(self, *pos, **kwargs):
        logger.debug("Evaluating definition: '%s'", self.__class__.__name__)
        logger.debug("--> Positional Args: %s", pos)
        logger.debug("--> Keyword arguments: %s", kwargs)

        # Requiring inputs to be given as kw args makes the api less sensitive to
        # changes in the implementation
        if len(pos) != 0:
            raise TypeError("Definition inputs must be passed as keyword arguments")

        required = self.inputs(inherited=False)
        missing = ["'" + n + "'" for n in required.keys() if n not in kwargs]

        logger.debug("Preparing arguments")
        logger.debug("--> Directly required inputs: %s", list(required.keys()))

        if len(missing) != 0:
            name = self.__class__.__name__
            inpts = ", ".join(missing)
            message = f"Unable to evaluate definition '{name}', missing inputs: {inpts}"

            raise TypeError(message)

        # Start building a dict for the args to pass to the _impl function. Starting
        # with the actual values of the required inputs
        args = {inpt: kwargs[inpt] for inpt in required}

        # Now to evaluate any definitions this definition is derived from.
        attributes = self.attributes(inherited=True)
        bases = dict(self._bases)

        logger.debug("--> Attributes: %s", attributes)
        logger.debug("--> Bases: %s", bases)

        for name, defn in bases.items():

            if name in kwargs:
                logger.debug("%s: Using user provided override, %s", name, kwargs[name])
                args[name] = kwargs[name]
                continue

            attrs = {name: attributes[name] for name in defn.attribs(inherited=True)}
            instance = defn(**attrs)
            logger.debug("%s: Created instance, %s", name, instance)

            # And then evaluate it!
            args[name] = instance(**kwargs)

        message = "Executing implementation for '%s' with args: %s"
        logger.debug(message, self.__class__.__name__, args)

        return self._impl(**args, **self.attributes())

    def _special_method(self, operation, a, b):
        """Implements the special methods in a standardized way."""

        op_name = operation.capitalize().replace("_", " ")
        a_is_defn = isinstance(a, Defn)
        b_is_defn = isinstance(b, Defn)

        t1 = type(a) if not a_is_defn else a.produces()
        t2 = type(b) if not b_is_defn else b.produces()

        impl = self._operators.get((operation, t1, t2), None)

        if impl is None:
            a = _format_type(a)
            b = _format_type(b)

            raise TypeError(f"{op_name} is not supported between {a} and {b}")

        defn = impl(a=a, b=b)
        return defn

    def __add__(self, other):
        return self._special_method(self.OP_ADD, self, other)

    def __radd__(self, other):
        return self._special_method(self.OP_ADD, other, self)

    def __and__(self, other):
        return self._special_method(self.OP_AND, self, other)

    def __rand__(self, other):
        return self._special_method(self.OP_AND, other, self)

    def __floordiv__(self, other):
        return self._special_method(self.OP_FLOORDIV, self, other)

    def __rfloordiv__(self, other):
        return self._special_method(self.OP_FLOORDIV, other, self)

    def __lshift__(self, other):
        return self._special_method(self.OP_LSHIFT, self, other)

    def __rlshift__(self, other):
        return self._special_method(self.OP_LSHIFT, other, self)

    def __matmul__(self, other):
        return self._special_method(self.OP_MATMUL, self, other)

    def __rmatmul__(self, other):
        return self._special_method(self.OP_MATMUL, other, self)

    def __mod__(self, other):
        return self._special_method(self.OP_MOD, self, other)

    def __rmod__(self, other):
        return self._special_method(self.OP_MOD, other, self)

    def __mul__(self, other):
        return self._special_method(self.OP_MUL, self, other)

    def __rmul__(self, other):
        return self._special_method(self.OP_MUL, other, self)

    def __or__(self, other):
        return self._special_method(self.OP_OR, self, other)

    def __ror__(self, other):
        return self._special_method(self.OP_OR, other, self)

    def __pow__(self, other):
        return self._special_method(self.OP_POW, self, other)

    def __rpow__(self, other):
        return self._special_method(self.OP_POW, other, self)

    def __rshift__(self, other):
        return self._special_method(self.OP_RSHIFT, self, other)

    def __rrshift__(self, other):
        return self._special_method(self.OP_RSHIFT, other, self)

    def __sub__(self, other):
        return self._special_method(self.OP_SUB, self, other)

    def __rsub__(self, other):
        return self._special_method(self.OP_SUB, other, self)

    def __truediv__(self, other):
        return self._special_method(self.OP_DIV, self, other)

    def __rtruediv__(self, other):
        return self._special_method(self.OP_DIV, other, self)

    def __xor__(self, other):
        return self._special_method(self.OP_XOR, self, other)

    def __rxor__(self, other):
        return self._special_method(self.OP_XOR, other, self)

    def attributes(self, inherited=False):
        """Return all attributes and their values on this definition instance.

        Parameters
        ----------
        inherited:
            If :code:`True` return all inherited attributes also.
        """

        if inherited:
            return {
                a.name: getattr(self, a.name)
                for a in attr.fields(self.__class__)
                if Defn.ATTR_ID in a.metadata
            }

        return {
            a.name: getattr(self, a.name)
            for a in attr.fields(self.__class__)
            if not a.metadata[Defn.ATTR_ID]["inherited"]
        }

    @classmethod
    def attribs(cls, inherited=False):
        """Return all attributes defined on this definition.

        Parameters
        ----------
        inherited:
            If :code:`True` return all inherited attributes also.
        """

        if inherited:
            return {a.name: a for a in attr.fields(cls) if Defn.ATTR_ID in a.metadata}

        return {
            a.name: a
            for a in attr.fields(cls)
            if not a.metadata[Defn.ATTR_ID]["inherited"]
        }

    @classmethod
    def bases(cls):
        """Return all the definitions this defintion is derived from."""

        if not hasattr(cls, "_bases"):
            return {}

        return dict(cls._bases)

    @classmethod
    def inputs(cls, inherited=True):
        """Return all the inputs required to evaluate this definition.

        Parameters
        ----------
        inherited:
            If :code:`True` (default) also return any inherited inputs.
        """
        if not hasattr(cls, "_inputs"):
            return {}

        if inherited:
            return {k: v for k, v in cls._inputs.items()}

        return {k: v for k, v in cls._inputs.items() if not v.inherited}

    @classmethod
    def produces(cls):
        """Return the type of the object that this definition produces."""
        rtype = inspect.signature(cls._impl).return_annotation

        if rtype == inspect._empty:
            return Any

        return rtype


@attr.s(auto_attribs=True)
class DefnInput:
    """A class that represents an input to a definition"""

    name: str
    """The name of the input."""

    dtype: Any
    """The type of the input"""

    inherited: bool
    """Flag to indicate if the input has been inherited from another definition."""

    sources: Optional[List[Defn]] = None
    """List of definitions to indicate which definitions an input has been inherited
    from."""


def _define_attribute(param: inspect.Parameter) -> attr.Attribute:
    """Given a parameter that represents some definition's attribute, write the
    corresponding attrs instance."""

    args = {"default": param.default, "kw_only": True}
    args["metadata"] = {Defn.ATTR_ID: {"inherited": False}}

    if param.annotation != inspect.Parameter.empty:
        args["type"] = param.annotation

        # TODO: Something, something validation...
        # args["validator"] = ...

    return attr.ib(**args)


def _inherit_attributes(defn: Defn, attributes):
    """Given a definition and the attributes for the current definition under
    construction copy over its attributes."""

    # Only look at the fields that represent attributes.
    for attrib in (a for a in attr.fields(defn) if a.metadata[Defn.ATTR_ID]):

        # For now we will skip any attributes that have already been defined.
        # This means that any definitions that share the same attribute name will
        # inherit the same value.
        #
        # This is probably what you would want in most(?) cases, but I'm sure we will
        # need the ability to do something more clever in the future...
        if attrib.name in attributes:
            continue

        metadata = dict(**attrib.metadata)
        metadata[Defn.ATTR_ID] = {"inherited": True}

        # It seems that the best way to copy an attrs attribute from one class to
        # another is to construct a fresh instance, copying over all the fields from
        # the source...
        attributes[attrib.name] = attr.ib(
            converter=attrib.converter,
            default=attrib.default,
            kw_only=attrib.kw_only,
            repr=False,
            metadata=metadata,
            validator=attrib.validator,
        )


def _inherit_inputs(defn: Defn, inputs):
    """Given a definition and the inputs for the current definition under construction
    copy over its inputs.

    Parameters
    ----------
    defn:
        The definition to inherit inputs from
    inputs:
        The dictionary carrying the inputs for the definition under construction.
    """
    logger.debug("Inherting inputs from definition: %s", defn.__name__)

    for name, inpt in defn.inputs().items():

        dname = defn.__name__

        if name not in inputs:
            inputs[name] = DefnInput(
                name=name, dtype=inpt.dtype, inherited=True, sources=[defn]
            )

            logger.debug("%s: inherited input, new", name)
            continue

        existing = inputs[name]

        # Check for a conflict between the types
        if existing.dtype != inpt.dtype:
            message = (
                f"Input '{name}' ({inpt.dtype}) inherited from '{dname}' conflicts"
                f" with existing input '{name}' ({existing.dtype})"
            )

            if existing.sources is not None:
                extras = ",".join(["'" + d.__name__ + "'" for d in existing.sources])
                message += f" inherited from {extras}"

            raise TypeError(message)

        # If the input is an inherited one, add this definition to the list of sources
        if existing.inherited:
            existing.sources.append(defn)
            logger.debug("%s: inherited input, updated", name)

            continue

        logger.debug("%s: already defined", name)


def _process_parameters(
    signature: inspect.Signature, attributes: Dict[str, Any]
) -> List[inspect.Parameter]:
    """Ensure that the input parameters for the defintion are well defined.

    All parameters must carry a type annotation, if the annotation is not a
    Defn the parameter will be registered as an input that must be given when
    evaluating a definiton.

    Otherwise if the annotation is another definition then both the inputs and
    attributes of that definition will be inherited.
    """

    # Input parameters are any arguments that are not attributes
    params = [p for p in signature.values() if p.name not in attributes]

    # We will hold a dict of definitions we are based on.
    defns = {}

    # As well as a dict of inputs required to evaluate the definiton.
    inputs = {}

    for param in params:

        if param.annotation == inspect.Parameter.empty:
            raise TypeError(f"Missing type annotation for parameter '{param.name}'")

        if issubclass(param.annotation, Defn):
            defns[param.name] = param.annotation
            continue

        inpt = DefnInput(name=param.name, dtype=param.annotation, inherited=False)
        inputs[inpt.name] = inpt

    # Now for each definition, inherit its attributes and inputs.
    for defn in defns.values():
        _inherit_attributes(defn, attributes)
        _inherit_inputs(defn, inputs)

    return defns, inputs


_OPERATOR_POOL = {}


def _define_operator(defn: Defn, operation: str, operator_pool):
    """Given a definition, check to see if it matches the criteria to be an operator."""

    attrs = defn.attribs(inherited=True)

    if "a" not in attrs.keys() or "b" not in attrs.keys():
        message = "Operators must define 2 attributes 'a' and 'b'"
        raise TypeError(message)

    a = attrs["a"].type
    b = attrs["b"].type

    if a is None:
        raise TypeError("Operator input 'a' is missing a valid type annotation")

    if b is None:
        raise TypeError("Operator input 'b' is missing a valid type annotation")

    if isinstance(a, DefnSignature):
        a = a.produces

    if isinstance(b, DefnSignature):
        b = b.produces

    key = (operation, a, b)

    if key in operator_pool:
        raise TypeError(
            f"A {operation} operator between {a} and {b} has already been defined"
        )

    operator_pool[key] = defn


def definition(f=None, *, operation: str = None, operator_pool=None):
    """Create a new Definition.

    Parameters
    ----------
    f:
        The function that is the definition
    operator:
        Flag used to indicate if this definition is an operator.
    operator_pool:
        Can be used to override the default operator pool
    """

    def wrapper(fn):

        name = fn.__name__
        operators = operator_pool if operator_pool is not None else _OPERATOR_POOL

        KW_ONLY = inspect.Parameter.KEYWORD_ONLY
        signature = inspect.signature(fn).parameters

        attrs = [attr for attr in signature.values() if attr.kind == KW_ONLY]

        attributes = {
            "__doc__": inspect.getdoc(fn),
            "__module__": fn.__module__,
            "_impl": staticmethod(fn),
            "_operators": operators,
        }

        for a in attrs:
            attributes[a.name] = _define_attribute(a)

        bases, inputs = _process_parameters(signature, attributes)
        attributes["_bases"] = bases
        attributes["_inputs"] = inputs

        defn = attr.s(type(name, (Defn,), attributes))

        if operation is not None:
            _define_operator(defn, operation, operators)

        return defn

    # Allow the decorator to be used with or without a function call
    if f is None:
        return wrapper

    return wrapper(f)



from ._core import Defn, DefnInput, Mask, definition  # noqa: F401
from ._expressions import all, any, clamp, invert, lerp, normalise  # noqa: F401
from ._images import Image, Resolutions, colorramp, encode, fill, save  # noqa: F401
from ._version import __version__  # noqa: F401



import os
import re
import textwrap

from pathlib import Path
from typing import Iterable, List, Set, Union

import docutils.nodes as nodes
import docutils.writers as writers
import nbformat.v4 as nbf

from docutils.io import StringOutput
from sphinx.builders import Builder
from sphinx.util import logging

from .directives import nbsolution, nbtutorial

logger = logging.getLogger(__name__)


class NotebookWriter(writers.Writer):
    """A class that walks a doctree converting it into a notebook."""

    def __init__(self, builder):
        super().__init__()
        self.builder = builder

    def translate(self):

        visitor = NotebookTranslator(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.astext()


class PythonTranslator(nodes.NodeVisitor):
    """Coverts an rst doctree into a valid python file.
    Currently only used to write tutorial solutions to a python file that can be
    imported via the `%load` magic in a jupyter notebook.
    """

    def __init__(self, document):
        super().__init__(document)
        self.is_code = False
        self.source = ""

    def astext(self):
        return self.source

    # ------------------------- Visitors --------------------------------------

    def visit_Text(self, node: nodes.Text) -> None:
        source = node.astext()
        self.source += textwrap.indent(source, "# " * self.is_code)

    def depart_Text(self, node: nodes.Text) -> None:
        pass

    def unknown_visit(self, node: nodes.Node) -> None:
        pass

    def unknown_departure(self, node: nodes.Node) -> None:
        pass


class NotebookTranslator(nodes.NodeVisitor):
    """Converts rst nodes into their output representation."""

    def __init__(self, document):
        super().__init__(document)
        self.cells = []
        self.level = 0
        self.prefix = None

    def asnotebook(self):
        return nbf.new_notebook(cells=self.cells)

    @property
    def current_cell(self):

        if len(self.cells) == 0:
            return None

        return self.cells[-1]

    def new_cell(self, cell_type: str) -> None:
        current = self.current_cell

        if current is not None and current.cell_type == cell_type:
            return

        types = {"markdown": nbf.new_markdown_cell, "code": nbf.new_code_cell}
        new_cell = types[cell_type]
        self.cells.append(new_cell())

    def append(self, text):

        if self.prefix is not None:
            text = text.replace("\n", "\n" + self.prefix)

            self.current_cell.source += text
            return

        self.current_cell.source += text

    def astext(self):
        notebook = nbf.new_notebook(cells=self.cells)
        return nbf.writes(notebook)

    # --------------------------- Visitors ------------------------------------

    def visit_bullet_list(self, node: nodes.bullet_list) -> None:
        self.new_cell("markdown")

    def depart_bullet_list(self, node: nodes.bullet_list) -> None:
        pass

    def visit_compound(self, node: nodes.compound) -> None:
        pass

    def depart_compound(self, node: nodes.compound) -> None:
        pass

    def visit_compact_paragraph(self, node) -> None:
        pass

    def depart_compact_paragraph(self, node) -> None:
        pass

    def visit_comment(self, node: nodes.comment) -> None:
        self.new_cell("markdown")

    def depart_comment(self, node: nodes.comment) -> None:
        pass

    def visit_document(self, node: nodes.document) -> None:
        pass

    def depart_document(self, node: nodes.document) -> None:
        pass

    def visit_emphasis(self, node: nodes.emphasis) -> None:
        self.current_cell.source += "*"

    def depart_emphasis(self, node: nodes.emphasis) -> None:
        self.current_cell.source += "*"

    def visit_inline(self, node: nodes.inline) -> None:
        self.current_cell.source += "["

    def depart_inline(self, node: nodes.inline) -> None:
        self.current_cell.source += "]"

    def visit_list_item(self, node: nodes.list_item) -> None:
        self.append("- ")

    def depart_list_item(self, node: nodes.list_item) -> None:
        pass

    def visit_literal(self, node: nodes.literal) -> None:
        self.current_cell.source += "`"

    def depart_literal(self, node: nodes.literal) -> None:
        self.current_cell.source += "`"

    def visit_literal_block(self, node: nodes.literal_block) -> None:
        self.new_cell("code")

    def depart_literal_block(self, node: nodes.literal_block) -> None:
        pass

    def visit_nbtutorial(self, node: nbtutorial) -> None:
        pass

    def depart_nbtutorial(self, node: nbtutorial) -> None:
        pass

    def visit_note(self, node: nodes.note) -> None:
        self.new_cell("markdown")
        self.append("> **Note**\n")
        self.prefix = "> "

    def depart_note(self, node: nodes.note) -> None:
        self.prefix = None

    def visit_paragraph(self, node: nodes.paragraph) -> None:
        self.new_cell("markdown")

        if isinstance(node.parent, nodes.list_item):
            return

        self.append("\n")

    def depart_paragraph(self, node: nodes.paragraph) -> None:
        self.append("\n")

    def visit_reference(self, node: nodes.reference) -> None:
        self.append("[")

    def depart_reference(self, node: nodes.reference) -> None:
        attrs = node.attributes

        url = attrs["refuri"]
        self.append(f"]({url})")

    def visit_section(self, node: nodes.section) -> None:
        self.level += 1
        self.new_cell("markdown")

    def depart_section(self, node: nodes.section) -> None:
        self.level -= 1

    def visit_strong(self, node: nodes.strong) -> None:
        self.current_cell.source += "**"

    def depart_strong(self, node: nodes.strong) -> None:
        self.current_cell.source += "**"

    def visit_target(self, node: nodes.target) -> None:
        pass

    def depart_target(self, node: nodes.target) -> None:
        pass

    def visit_Text(self, node: nodes.Text) -> None:

        if isinstance(node.parent, nodes.comment):
            return

        if self.current_cell.cell_type == "markdown":
            self.append(node.astext())
            return

        source = node.astext()

        if ">>>" not in source:
            self.append(source)
            return

        pattern = re.compile("^(>>>|\\.\\.\\.) ?")

        def clean_line(line):
            return pattern.sub("", line)

        cleaned_source = "\n".join(
            [clean_line(line) for line in source.split("\n") if pattern.match(line)]
        )

        self.append(cleaned_source)

    def depart_Text(self, node: nodes.Text) -> None:
        pass

    def visit_title(self, node: nodes.title) -> None:
        title = "#" * self.level
        self.append(f"{title} ")

    def depart_title(self, node: nodes.title) -> None:
        self.append("\n")


def codeblock(source: str) -> nodes.literal_block:
    """Construct a valid code block."""

    block = nodes.literal_block()
    block.children = [nodes.Text(source)]

    return block


class NotebookTutorialBuilder(Builder):
    """Builder that can convert static tutorials into an interactive jupyer
    notebook."""

    name = "nbtutorial"
    format = "ipynb"

    def init(self) -> None:
        """Any initialization goes here."""
        logger.debug(f"[nbtutorial]: Outdir is: {self.outdir}")

    def get_outdated_docs(self) -> Union[str, Iterable[str]]:
        """Not too sure what we should do here yet."""

        return self.env.found_docs

    def get_target_uri(self, docname: str, type: str = None) -> str:
        """Another method to figure out."""

        uri = docname + ".ipynb"

        logger.debug(f"[nbtutorial]: Target URI: {uri}")

        return uri

    def _process_solutions(self, docname: str, solutions: List[nbsolution]) -> None:
        """Given the solutions for a given tutorial save them to the solutions dir.

        This also rewrites the doctree so that the solutions are replaced by cells
        the :code:`%load` magic so that the user can load the results in.
        """
        logger.debug("[nbtutorial]: Processing solutions for %s", docname)

        DIRNAME = "solutions"
        soln_dir = os.path.join(self.outdir, os.path.dirname(docname), DIRNAME)
        soln_name = os.path.basename(docname)

        if not os.path.exists(soln_dir):
            os.makedirs(soln_dir)

        for idx, soln in enumerate(solutions):
            soln_fname = f"{soln_name}-{idx + 1:02d}.py"
            soln_path = os.path.join(DIRNAME, soln_fname)

            # Convert the solution to a valid Python file
            translator = PythonTranslator(soln)
            soln.walkabout(translator)
            python_soln = translator.astext()

            # Insert a code block into the notebook that will load the solution
            soln.children = [codeblock(f"%load {soln_path}")]

            # Write the actual solution to the given file on disk
            soln_file = os.path.join(soln_dir, soln_fname)
            logger.debug("[nbtutorial]: --> %s", soln_path)

            with open(soln_file, "w") as f:
                f.write(python_soln)

    def prepare_writing(self, docnames: Set[str]) -> None:
        """A place we can add logic to?"""

        self.docwriter = NotebookWriter(self)

    def write_doc(self, docname: str, doctree: nodes.Node) -> None:
        logger.debug(f"[nbtutorial]: Called on {docname}")

        # Determine if the document represents a tutorial.
        nodes = list(doctree.traverse(condition=nbtutorial))

        if len(nodes) == 0:
            return

        # Find any solutions that it may constain.
        solutions = list(doctree.traverse(condition=nbsolution))
        self._process_solutions(docname, solutions)

        destination = StringOutput(encoding="utf-8")
        self.docwriter.write(doctree, destination)

        base, fname = os.path.split(docname)
        basedir = os.path.join(self.outdir, base)

        if not os.path.exists(basedir):
            os.makedirs(basedir)

        Path(basedir, "__init__.py").touch()
        outfile = os.path.join(basedir, fname + ".ipynb")

        with open(outfile, "w") as f:
            f.write(self.docwriter.output)



import logging

from typing import Any

import attr

try:
    import tkinter as tk
    from PIL import ImageTk
except ImportError:
    import unittest.mock as mock

    tk = mock.MagicMock()


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class CanvasItem:

    source: Any = None
    """Represents the source of truth for the item"""

    tk: Any = None
    """The tk representation of the item"""

    canvas: Any = None
    """The canvas representation of the item"""


class ArlunioCanvas(tk.Canvas):
    """A canvas object, capable of display arlunio related items"""

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, highlightthickness=0, **kwargs)
        self.bind("<Configure>", self.on_resize)

        self._item = None
        self.height = self.winfo_reqheight()
        self.width = self.winfo_reqwidth()

    def on_resize(self, event):
        self.width = event.width
        self.height = event.height

        logger.debug("Resizing canvas %s x %s", self.width, self.height)
        self.config(width=self.width, height=self.height)

        cx, cy = self.width / 2, self.height / 2

        img = self._item.source(self.width, self.height)._as_pillow_image()
        self._item.tk = ImageTk.PhotoImage(img)
        self.coords(self._item.canvas, cx, cy)
        self.itemconfig(self._item.canvas, image=self._item.tk)

    def set_item(self, item):
        canvas_item = CanvasItem(source=item)

        img = item(self.width, self.height)._as_pillow_image()
        canvas_item.tk = ImageTk.PhotoImage(img)

        cx, cy = self.width / 2, self.height / 2
        canvas_item.canvas = self.create_image(cx, cy, image=canvas_item.tk)
        self._item = canvas_item


class ImageViewer(tk.Frame):
    """A component dedicated to displaying images"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.pack(fill=tk.BOTH, expand=tk.TRUE)

        self.canvas = ArlunioCanvas(self, width=640, height=480)
        self.canvas.pack(fill=tk.BOTH, expand=tk.TRUE)

    def set_item(self, item):
        self.canvas.set_item(item)
